using System.Net;
using System.Reflection;
using System.Security.Cryptography;
using System.Text;
using Autofac;
using Autofac.Core.Activators.Reflection;
using FitConnect.Encryption;
using FitConnect.Interfaces.Subscriber;
using FitConnect.Models;
using FitConnect.Models.v1.Api;
using FitConnect.Services.Models.v1.Submission;
using IdentityModel;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using NJsonSchema;
using NJsonSchema.Validation;
using Metadata = FitConnect.Models.Api.Metadata.Metadata;

namespace FitConnect;

/// <summary>
///     Fluent API for the FitConnect.Subscriber
/// </summary>
public class Subscriber : FitConnectClient,
    ISubscriber,
    ISubscriberWithSubmission {
    
    public Submission? Submission { get; private set; }
    
    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        string publicKeyEncryption,
        string publicKeySignatureVerification,
        IContainer container) : base(environment, clientId, clientSecret,
        container,
        privateKeyDecryption, privateKeySigning, publicKeyEncryption,
        publicKeySignatureVerification) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning, publicKeyEncryption,
            publicKeySignatureVerification,
            container.Resolve<ILogger>());
    }

    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        string publicKeyEncryption,
        string publicKeySignatureVerification,
        ILogger? logger = null) : base(environment, clientId, clientSecret, logger,
        privateKeyDecryption, privateKeySigning) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning, publicKeyEncryption,
            publicKeySignatureVerification, logger);
    }


    /// <summary>
    ///     Receives the available submissions from the server
    /// </summary>
    /// <param name="destinationId"></param>
    /// <param name="skip"></param>
    /// <param name="take"></param>
    /// <returns></returns>
    public IEnumerable<SubmissionForPickupDto> GetAvailableSubmissions(string? destinationId = null,
        int skip = 0,
        int take = 100) {
        var submissionsResult = SubmissionService.ListSubmissions(destinationId, 0, 100);

        // Creating a dictionary of destinationId to submissionIds from the REST API result
        return submissionsResult.Submissions ?? new List<SubmissionForPickupDto>();
    }


    /// <summary>
    ///     Receives a specific submission from the server.
    /// </summary>
    /// <param name="submissionId"></param>
    /// <param name="skipSchemaTest"></param>
    /// <returns></returns>
    public ISubscriberWithSubmission RequestSubmission(string submissionId,
        bool skipSchemaTest = false) {
        var submission = (Submission)SubmissionService.GetSubmission(submissionId);
        var (metaDataString, _, metaHash) = Encryption.Decrypt(submission.EncryptedMetadata);

        if (!skipSchemaTest) {
            var valid = JsonHelper.VerifyMetadata(metaDataString);
            if (!valid) {
                Logger?.LogWarning("Invalid metadata: {MetaData}", metaDataString);

                throw new Exception($"Metadata validation failed");
            }
        }

        submission.Metadata = JsonConvert.DeserializeObject<Metadata>(metaDataString);


        if (submission.EncryptedData != null) {
            var (dataString, _, dataHash) = Encryption.Decrypt(submission.EncryptedData);
            submission.Data = dataString;

            if (submission.Metadata?.ContentStructure.Data.Hash.Content !=
                FitEncryption.CalculateHash(dataString)) {
                Logger?.LogWarning("Data hash mismatch: {DataHash} != {CalculatedHash}",
                    submission.Metadata?.ContentStructure.Data.Hash.Content,
                    FitEncryption.CalculateHash(dataString));
                throw new Exception("Data hash mismatch");
            }
        }

        Submission = submission;
        return this;
    }

    

    /// <summary>
    ///     Reading attachments for a submission.
    /// </summary>
    /// <returns></returns>
    public IEnumerable<Attachment> GetAttachments() {
        if (Submission?.Id == null || Submission?.Metadata == null)
            throw new Exception("No submission available");

        var attachments = new List<Attachment>();
        foreach (var id in Submission!.AttachmentIds) {
            var encryptedAttachment = SubmissionService.GetAttachment(Submission.Id, id);
            var (_, content, hash) = Encryption.Decrypt(encryptedAttachment);
            var attachmentMeta =
                Submission.Metadata.ContentStructure.Attachments.First(a => a.AttachmentId == id);

            attachments.Add(new Attachment(attachmentMeta, content));
        }

        Submission.Attachments = attachments;
        return attachments;
    }

    public void AcceptSubmission() {
        CompleteSubmission(Submission!, FinishSubmissionStatus.Accepted);
    }

    public void RejectSubmission(params Problems[] problems) {
        CompleteSubmission(Submission!, FinishSubmissionStatus.Rejected, problems);
    }


    // public void ForwardSubmission() {
    //     CompleteSubmission(Submission!, FinishSubmissionStatus.Forwarded);
    // }

    public void CompleteSubmission(FinishSubmissionStatus status) {
        CompleteSubmission(Submission!, status);
    }



    private void CompleteSubmission(SubmissionForPickupDto submission,
        FinishSubmissionStatus status, Problems[]? problems = null) {
        if (submission.SubmissionId == null || submission.CaseId == null ||
            submission.DestinationId == null)
            throw new ArgumentException("Submission does not contain all required fields");

        if (status != FinishSubmissionStatus.Rejected && problems != null)
            throw new ArgumentException("Problems can only be set for rejected submissions");

        var token = status switch {
            FinishSubmissionStatus.Rejected =>
                Encryption.CreateRejectSecurityEventToken(submission.SubmissionId,
                    submission.CaseId,
                    submission.DestinationId, problems),
            FinishSubmissionStatus.Accepted => Encryption.CreateAcceptSecurityEventToken(
                submission),
            _ => throw new ArgumentOutOfRangeException(nameof(status), status, null)
        };

        Logger?.LogDebug("Token to accept submission: {Token}", token);
        var result = CasesService.FinishSubmission(submission.CaseId, token);
        Logger?.LogInformation("Submission completed {Status}", result);
    }

    public static string VerifyCallback(string callbackSecret,
        long timestamp, string body) {
        if (timestamp < DateTime.Now.AddMinutes(-5).ToEpochTime())
            throw new ArgumentException("Request is too old");

        var hmac = new HMACSHA512(Encoding.UTF8.GetBytes(callbackSecret))
            .ComputeHash(Encoding.UTF8.GetBytes($"{timestamp}.{body}"));


        return Convert.ToHexString(hmac).ToLower();
    }

    public static bool VerifyCallback(string callbackSecret, HttpRequest request) {
        if (!request.Headers.ContainsKey("callback-timestamp"))
            throw new ArgumentException("Missing callback-timestamp header");

        var timeStampString = request.Headers["callback-timestamp"].ToString();
        if (!long.TryParse(timeStampString, out var timestamp)) {
            throw new ArgumentException("Invalid callback-timestamp header");
        }


        var authentication = request.Headers["callback-authentication"];
        using var requestStream = request.Body;
        var content = new StreamReader(requestStream).ReadToEnd().Trim();

        var result = VerifyCallback(callbackSecret, timestamp, content);
        if (result != authentication)
            throw new ArgumentException("Request is not authentic");
        return true;
    }
}

public enum FinishSubmissionStatus {
    Accepted,

    Rejected
    // Forwarded
}
